
#ifndef H_TESLA_INTERNAL
#define H_TESLA_INTERNAL

#include <stdio.h>
#include <sys/socket.h>

#define ADDRESS_SIZE 256
#define DATA_BUFFER_SIZE 8192
#define MAX_DGRAM_SIZE 65536
#define MAX_OPTVAL_SIZE 1024
/* assert(MAX_DGRAM_SIZE > DATA_BUFFER_SIZE) */

#ifdef __cplusplus
extern "C" {
#endif

// Structure to track information about a single socket.

typedef struct
{
    int wrapped;  // Is it wrapped?
    int domain;   // e.g., PF_INET
    int type;     // e.g., SOCK_DGRAM
    int conn_id;  // connection ID used when communicating with master
    int connecting;
    int connect_status;
} ts_fd_info;
extern ts_fd_info ts_fds[FD_SETSIZE];

typedef struct
{
    size_t    length;
    socklen_t addrlen;
    struct sockaddr addr;
} ts_dgram_header;

// Stuff about application<->master communication.

typedef enum {
    MSG_HELLO,
    MSG_SOCKET,
    MSG_CONNECT,
    MSG_FH_FOLLOWS,
    MSG_ACK,
    MSG_NAK,
    MSG_BIND,
    MSG_LISTEN,
    MSG_ACCEPT,
    MSG_GETSOCKOPT,
    MSG_SETSOCKOPT,
    MSG_GETPEERNAME,
    MSG_GETSOCKNAME,
    MSG_ASYNC_MESSAGES,
    MSG_FORK,
    MSG_INESSENTIAL,

    MSG_TYPE_COUNT
} master_msg_type_t;

extern const char *__ts_msg_types[];

#define MSG_TYPE(i) ((i) >= 0 && (i) < MSG_TYPE_COUNT ? \
                     __ts_msg_types[(i)] : "UNKNOWN")

typedef struct {
    char handler[32];
    unsigned int conn_id;
    int name;
    int data_length;
} ts_async_message_s;

typedef struct {
    size_t len;
    unsigned int id;
    unsigned int conn_id;
    pid_t pid;
    unsigned int merrno;
    master_msg_type_t type;

    union {
	struct {
	} hello;
	struct {
	    int domain, type, protocol;
	} socket;
	struct {
	    int fh;
	} fh_pass;
	struct {
	    int addrlen;
	    char addr[ADDRESS_SIZE]; /* XXX: Copout */
	} address;
	struct {
	    int level;
	    int optname;
	    int optlen;
	    char target[32];
	    char optval[1024]; /* XXX: Also a copout */
	} sockopt;	    
	struct {
	    int enabled;
	    int pass;
	} async_messages;
	struct {
	    int val;
	} inessential;
    } body;
} master_msg_t;

#define OFFSET(str, mem) (((char*)&(((str*)0)->mem)) - (char*)0)
#define OFFSET_FROM(str, mem1, mem2) (OFFSET(str, mem2) - OFFSET(str, mem1))
#define MASTER_MSG_HDRLEN OFFSET(master_msg_t, body)



// Macros to call libc functions directly.

extern void *__ts_libc;
#define TS_CALL_LIBC(func, args...) \
({ \
    static typeof(func) *f = NULL;                                             \
    if (!__ts_libc) { __ts_libc = dlopen("libc.so.6", RTLD_LAZY); if (!__ts_libc) exit(1); } \
    if (!f) { f = dlsym(__ts_libc, #func); if (!f) exit(1); }                      \
    (*f)(args); \
})

#define MAX_APPLY_ARGS 6
#define TS_APPLY_LIBC(func) do { \
    static void (*f)() = NULL;                                             \
    if (!__ts_libc) { __ts_libc = dlopen("libc.so.6", RTLD_LAZY); if (!__ts_libc) exit(1); }  \
    if (!f) { f = dlsym(__ts_libc, #func); if (!f) exit(1); }                       \
    __builtin_return(__builtin_apply(f, __builtin_apply_args(), MAX_APPLY_ARGS * sizeof(int))); \
} while (0)
#define APPLY_DEFAULT(func) __builtin_return(__builtin_apply(default_##func, __builtin_apply_args(), MAX_APPLY_ARGS * sizeof(int)))

// Variables used for debugging
extern int __ts_debug_level;
extern FILE *__ts_debug_file;
extern void __ts_debug_init();

#ifdef __cplusplus
};
#endif

#endif
